/*
 * Common display-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __DISPLAY_H__
#define __DISPLAY_H__

#include <stdint.h>



/* Display configuration type. */

typedef struct
{
    /* Framebuffer address, potentially referencing many frames. */

    uint8_t *framebuffer;

    /* The total number of lines in the framebuffer including spare lines to
       support horizontal scrolling. */

    uint32_t total_lines;

    /* The screen dimensions and size (derived from the dimensions). */

    uint32_t line_length, line_count;   /* width, height */
    uint32_t screen_size;               /* = width * height */

    /* The current frame, number of frames, and the maximum number supported. */

    int frame, frames, max_frames;

    /* The fixed start address of the current frame. */

    uint8_t *frame_start;               /* = framebuffer + (frame_size * frame) */

    /* The floating screen start/top pointer. */

    uint8_t *screen_start;              /* >= frame_start; < screen_limit */

    /* A pointer to a collection of screen start pointers, one per frame. */

    uint8_t **screen_starts;

    /* The fixed end address of the current frame. */

    uint8_t *screen_limit;              /* = frame_start + screen_size */

    /* Number of scanlines on the display. */

    uint32_t scanlines;

    /* Number of scanlines per display line. */

    int line_multiplier;                /* = scanlines / line_count */

    /* Number of line channels used for pixel data transfers. */

    int line_channels;

    /* Number of consecutive pixels provided by a framebuffer region. */

    uint32_t cell_size;

    /* DMA transfer sizes. */

    uint32_t transfer_cell_size;

    /* Display line positions. */

    uint32_t hfreq_limit, hsync_start, hsync_end;

    /* Display region scanline positions. */

    uint32_t visible_start, vfp_start, vsync_start, vsync_end;

} display_config_t;



/* Configuration functions. */

void init_display(display_config_t *cfg, uint8_t *framebuffer,
                  uint32_t line_length, uint32_t line_count,
                  int frames, uint8_t **screen_starts);

void init_display_properties(display_config_t *cfg);

void init_frames(display_config_t *cfg);

void display_select_frame(display_config_t *cfg, int frame);

void display_select_next_frame(display_config_t *cfg);

void display_set_frames(display_config_t *cfg, int frames);

uint8_t *display_get_frame_start(display_config_t *cfg, int frame);

uint32_t display_get_start_offset(display_config_t *cfg);

/* Access functions. */

int display_get_position(display_config_t *cfg, int x);

void display_copy(display_config_t *cfg, uint8_t *store,
                  int width, int height, int ystep,
                  int x, int y, int key, int to_display);

void display_copy_section(display_config_t *cfg, uint8_t *store,
                  int width, int height,
                  int xstart, int ystart, int xsize, int ysize, int ystep,
                  int x, int y, int key, int to_display);

void display_scroll(display_config_t *cfg, int x, int y);

void display_test_linedata(display_config_t *cfg);

/* Address-related functions. */

uint8_t *display_wrap_pointer(display_config_t *cfg, uint8_t *ptr);

#endif /* __DISPLAY_H__ */
