/*
 * Common viewport-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "viewport.h"



/* Initialise a viewport with display, origin, scaling and update function
   details. */

void init_viewport(viewport_t *v, display_config_t *cfg,
                   int xorigins[], int yorigins[],
                   int xscale, int yscale,
                   void (*update)())
{
    int frame;

    v->cfg = cfg;

    /* Set the storage for the frame origin details. */

    v->xorigins = xorigins;
    v->yorigins = yorigins;

    /* Set the relationship between origin coordinates and display columns and
       rows. */

    v->xscale = xscale;
    v->yscale = yscale;

    /* Set the post-scroll update function. */

    v->update = update;

    /* Set the origin to (0, 0) for every frame. */

    v->xorigin = 0;
    v->yorigin = 0;

    for (frame = 0; frame < cfg->frames; frame++)
    {
        v->xorigins[frame] = 0;
        v->yorigins[frame] = 0;
    }
}

/* Update the origin by the given changes in origin coordinates. */

void viewport_update_origin(viewport_t *v, int dx, int dy)
{
    int xstep, ystep;
    int frame = v->cfg->frame;

    /* Change the origin coordinates. */

    v->xorigin += dx;
    v->yorigin += dy;

    /* Determine the magnitude of the scrolling required from this frame to the
       current origin. */

    xstep = v->xorigin - v->xorigins[frame];
    ystep = v->yorigin - v->yorigins[frame];

    /* Scroll the frame, converting to columns and rows. */

    display_scroll(v->cfg, xstep / v->xscale, ystep / v->yscale);

    /* For horizontal scrolling, plot the exposed column at the left (if
       scrolling left) or at the right (if scrolling right). */

    if (v->update)
        v->update(v, v->xorigin, v->yorigin, xstep, ystep);

    /* Record the origin for this frame. */

    v->xorigins[frame] = v->xorigin;
    v->yorigins[frame] = v->yorigin;
}
