#!/usr/bin/env python

"""
Convert images for display in an Acorn Electron MODE 2 variant with a pixel
layout of I0RRGGBB, giving 128 colours instead of the usual 8 colours.

Copyright (C) 2015, 2017, 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from os.path import split, splitext
import EXIF
import PIL.Image
import sys

def convert_image(im, output_filename, width, height, strip, bytealign, options, label="screendata"):

    "Convert 'im' and write pixel values to 'output_filename'."

    w, h = im.size

    if strip:
        leftpad = rightpad = toppad = bottompad = 0
        width = w; height = h
    else:
        hpad = (width - w) / 2
        leftpad = hpad; rightpad = width - w - hpad
        vpad = (height - h) / 2
        toppad = vpad; bottompad = height - h - vpad

    data = iter(im.getdata())

    f = open(output_filename, "w")
    try:
        print >>f, """\
.section .rodata, "a"

/* Options:
%s
*/

.globl %s
.globl %s_width
.globl %s_height

%s:
%s_width:
.word %d
%s_height:
.word %d

%s_image:
""" % (options, label, label, label, label, label, width, label, height, label)

        word = []
        y = 0

        while y < height:
            x = 0

            # Top and bottom padding.

            if y < toppad or y >= height - bottompad:

                while x < width:
                    word.append(0)
                    flush_word(f, word, bytealign)
                    x += 1

                flush_last_word(f, word, bytealign)

            # Lines with data.

            else:
                while x < width:

                    # Left and right padding.

                    if x < leftpad or x >= width - rightpad:
                        word.append(0)

                    # Data regions.

                    else:
                        r, g, b = data.next()
                        rm, gm, bm, i = get_values(r, g, b)

                        # Encode the byte value: I0RRGGBB.

                        word.insert(0,
                            # I -> D<7>
                            (i << 7) |
                            # R<7:6> -> D<5:4>
                            (rm >> 2) |
                            # G<7:6> -> D<3:2>
                            (gm >> 4) |
                            # B<7:6> -> D<1:0>
                            (bm >> 6))

                    flush_word(f, word, bytealign)
                    x += 1

                flush_last_word(f, word, bytealign)

            y += 1

    finally:
        f.close()

def get_values(r, g, b):

    "Return modified values for 'r', 'g' and 'b', plus an intensity bit."

    rm = r & 0xc0
    gm = g & 0xc0
    bm = b & 0xc0
    rd = r - rm
    gd = g - gm
    bd = b - bm
    i = ((rd ** 2 + gd ** 2 + bd ** 2) ** 0.5) >= 32 and 1 or 0
    return rm, gm, bm, i

def make_preview(im):
    imp = PIL.Image.new("RGB", im.size)
    data = []
    for r, g, b in im.getdata():
        rm, gm, bm, i = get_values(r, g, b)
        r = rm + (i * 32)
        g = gm + (i * 32)
        b = bm + (i * 32)
        data.append((r, g, b))
    imp.putdata(data)
    return imp

def flush_last_word(f, word, bytealign):
    if word:
        if bytealign:
            write_bytes(f, word)
        else:
            pad_word(word)
            write_word(f, word)
        del word[:]

def flush_word(f, word, bytealign):
    if len(word) == 4:
        if bytealign:
            write_bytes(f, word)
        else:
            write_word(f, word)
        del word[:]

def pad_word(word):
    while len(word) < 4:
        word.insert(0, 0)

def write_bytes(f, word):
    while word:
        print >>f, ".byte 0x%02x" % word.pop()

def write_word(f, word):
    print >>f, ".word 0x%02x%02x%02x%02x" % tuple(word)

def rotate_and_scale(exif, im, base_width, base_height, rotate, width, height):

    """
    Using the given 'exif' information, rotate and scale image 'im' given the
    indicated 'base_width' and 'base_height' constraints and any explicit
    'rotate' indication.

    The 'width' and 'height' indicate the final dimensions.

    The returned image will be within the given 'width' and 'height', filling
    either or both, and preserve its original aspect ratio.
    """

    if rotate or exif and exif["Image Orientation"].values == [6L]:
        im = im.rotate(270)

    w, h = im.size

    # Get the relationship between the base width and the image width.

    wsf = float(base_width) / w

    # Get the relationship between the base height and the image height.

    hsf = float(base_height) / h

    # Determine the maximal scaling down required to fit the image.

    min_scale_factor = min(wsf, hsf)

    # Determine the final scaling factors to yield the desired dimensions.

    xscale = float(width) / base_width
    yscale = float(height) / base_height

    if min_scale_factor < 1:
        width = int(min_scale_factor * w * xscale)
        height = int(min_scale_factor * h * yscale)
        return im.resize((width, height))
    elif scale_factor != 1:
        width = int(w * wsf)
        return im.resize((width, h))
    else:
        return im

def get_parameter(options, flag, conversion, default, missing):

    """
    From 'options', return any parameter following the given 'flag', applying
    the 'conversion' which has the given 'default' if no valid parameter is
    found, or returning the given 'missing' value if the flag does not appear at
    all.
    """

    try:
        i = options.index(flag)
        try:
            return conversion(options[i+1])
        except (IndexError, ValueError):
            return default
    except ValueError:
        return missing

# Main program.

if __name__ == "__main__":

    # Test options.

    if "--help" in sys.argv or len(sys.argv) < 3:
        basename = split(sys.argv[0])[1]
        print >>sys.stderr, """\
Usage:

%s <input filename> <output filename> <label> [ <options> ]

Preview options:

-p - Generate a preview with a filename based on the output filename

Size options:

-W - Indicate the output width (default is 160)

-H - Indicate the output height (default is 256)

-S - Employ the width and height to define the appropriate output ratio but
     refine the final width and height and strip away padding

Transformation options:

-r - Rotate the input image clockwise explicitly
     (EXIF information is used otherwise)

Output options:

-b - Use bytes instead of padded words for each line of the output
""" % basename
        sys.exit(1)

    base_width = 320; width = 160
    base_height = height = 256

    input_filename, output_filename, label = sys.argv[1:4]
    options = sys.argv[3:]

    # Basic image properties.

    width = get_parameter(options, "-W", int, width, width)
    height = get_parameter(options, "-H", int, height, height)
    bytealign = "-b" in options
    rotate = "-r" in options
    preview = "-p" in options
    strip = "-S" in options

    # Load the input image.

    exif = EXIF.process_file(open(input_filename))
    im = PIL.Image.open(input_filename).convert("RGB")
    im = rotate_and_scale(exif, im, base_width, base_height, rotate,
                          width, height)

    # Generate an output image.

    convert_image(im, output_filename, width, height, strip, bytealign, options, label)

    # Generate a preview image if requested.

    if preview:
        _basename, ext = splitext(input_filename)
        basename, _ext = splitext(output_filename)
        preview_filename = "%s_preview%s" % (basename, ext)
        imp = make_preview(im)
        imp.save(preview_filename)

# vim: tabstop=4 expandtab shiftwidth=4
